/**
 * Copyright (C) 2015 Wasabeef
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
 "use strict";

var RE = {};

window.onload = function() {
    RE.callback("ready");
};

RE.editor = document.getElementById('editor');

// Not universally supported, but seems to work in iOS 7 and 8
document.addEventListener("selectionchange", function() {
    RE.backuprange();
});

// Not universally supported, but seems to work in iOS 7 and 8
document.addEventListener("click", function() {
    RE.callback("click");
});

//looks specifically for a Range selection and not a Caret selection
RE.rangeSelectionExists = function() {
    //!! coerces a null to bool
    var sel = document.getSelection();
    if (sel && sel.type == "Range") {
        return true;
    }
    return false;
};

RE.rangeOrCaretSelectionExists = function() {
    //!! coerces a null to bool
    var sel = document.getSelection();
    if (sel && (sel.type == "Range" || sel.type == "Caret")) {
        return true;
    }
    return false;
};

RE.editor.addEventListener("input", function() {
    RE.updatePlaceholder();
    RE.backuprange();
    RE.callback("input");
});

RE.editor.addEventListener("focus", function() {
    RE.backuprange();
    RE.callback("focus");
});

RE.editor.addEventListener("blur", function() {
    RE.callback("blur");
});

RE.customAction = function(action) {
    RE.callback("action/" + action);
};

RE.updateHeight = function() {
    RE.callback("updateHeight");
}

RE.callbackQueue = [];
RE.runCallbackQueue = function() {
    if (RE.callbackQueue.length === 0) {
        return;
    }

    setTimeout(function() {
        window.location.href = "re-callback://";
    }, 0);
};

RE.getCommandQueue = function() {
    var commands = JSON.stringify(RE.callbackQueue);
    RE.callbackQueue = [];
    return commands;
};

RE.callback = function(method) {
    RE.callbackQueue.push(method);
    RE.runCallbackQueue();
};

RE.setHtml = function(contents) {
    var tempWrapper = document.createElement('div');
    tempWrapper.innerHTML = contents;
    var images = tempWrapper.querySelectorAll("img");

    for (var i = 0; i < images.length; i++) {
        images[i].onload = RE.updateHeight;
    }

    RE.editor.innerHTML = tempWrapper.innerHTML;
    RE.updatePlaceholder();
};

RE.getHtml = function() {
    return RE.editor.innerHTML;
};

RE.getText = function() {
    return RE.editor.innerText;
};

RE.setPlaceholderText = function(text) {
    RE.editor.setAttribute("placeholder", text);
};

RE.updatePlaceholder = function() {
    if (RE.editor.textContent.length > 0) {
        RE.editor.classList.remove("placeholder");
    } else {
        RE.editor.classList.add("placeholder");
    }
};

RE.removeFormat = function() {
    document.execCommand('removeFormat', false, null);
};

RE.setFontSize = function(size) {
    RE.editor.style.fontSize = size;
};

RE.setBackgroundColor = function(color) {
    RE.editor.style.backgroundColor = color;
};

RE.setHeight = function(size) {
    RE.editor.style.height = size;
};

RE.undo = function() {
    document.execCommand('undo', false, null);
};

RE.redo = function() {
    document.execCommand('redo', false, null);
};

RE.setBold = function() {
    document.execCommand('bold', false, null);
};

RE.setItalic = function() {
    document.execCommand('italic', false, null);
};

RE.setSubscript = function() {
    document.execCommand('subscript', false, null);
};

RE.setSuperscript = function() {
    document.execCommand('superscript', false, null);
};

RE.setStrikeThrough = function() {
    document.execCommand('strikeThrough', false, null);
};

RE.setUnderline = function() {
    document.execCommand('underline', false, null);
};

RE.setTextColor = function(color) {
    RE.restorerange();
    document.execCommand("styleWithCSS", null, true);
    document.execCommand('foreColor', false, color);
    document.execCommand("styleWithCSS", null, false);
};

RE.setTextBackgroundColor = function(color) {
    RE.restorerange();
    document.execCommand("styleWithCSS", null, true);
    document.execCommand('hiliteColor', false, color);
    document.execCommand("styleWithCSS", null, false);
};

RE.setHeading = function(heading) {
    document.execCommand('formatBlock', false, '<h' + heading + '>');
};

RE.setIndent = function() {
    document.execCommand('indent', false, null);
};

RE.setOutdent = function() {
    document.execCommand('outdent', false, null);
};

RE.setOrderedList = function() {
    document.execCommand('insertOrderedList', false, null);
};

RE.setUnorderedList = function() {
    document.execCommand('insertUnorderedList', false, null);
};

RE.setJustifyLeft = function() {
    document.execCommand('justifyLeft', false, null);
};

RE.setJustifyCenter = function() {
    document.execCommand('justifyCenter', false, null);
};

RE.setJustifyRight = function() {
    document.execCommand('justifyRight', false, null);
};

RE.insertTextWithFormat = function(text) {
    RE.restorerange();
    
    const editorContent = RE.editor.innerText.trim();
    const insertText = editorContent !== "" ? "\n" + text : text;
    const sel = document.getSelection();

    const isListItemEmpty = (listItem) => {
        const text = listItem.textContent.trim();
        const listPrefixPattern = /^\d+\.\s*|\s*[-*]\s*/;
        return text.replace(listPrefixPattern, '').trim() === "";
    };
    
    const findLastListElement = (children) => {
        if (!children || children.length === 0) return null;

        let child = children[children.length - 1];
        if (child.nodeName.toLowerCase() === 'div') {
            let lastElementInDiv = child.lastElementChild;
            if (lastElementInDiv && (lastElementInDiv.nodeName === 'UL' || lastElementInDiv.nodeName === 'OL')) {
                return lastElementInDiv;
            }
            return findLastListElement(child.children);
        }
        return null;
    };

    if (sel.toString().length === 0) {
        const lastChild = RE.editor.lastElementChild;
        const isLastItemList = lastChild && (lastChild.nodeName === "UL" || lastChild.nodeName === "OL");
        
        RE.editor.focus();
        
        if (isLastItemList) {
            const lastListItem = lastChild.lastElementChild;
            if (lastListItem && isListItemEmpty(lastListItem)) {
                RE.insertTextWithFormatAtCurrentCursorPosition(text);
            } else {
                RE.insertTextWithFormatWithNewLine(text);
            }
        } else {
            const lastElementInDiv = findLastListElement(RE.editor.childNodes);
            if (lastElementInDiv) {
                const lastListItem = lastElementInDiv.lastElementChild;
                if (lastListItem && lastListItem.nodeName === 'LI') {
                    if (isListItemEmpty(lastListItem)) {
                        RE.insertTextWithFormatAtCurrentCursorPosition(text);
                    } else {
                        RE.insertTextWithFormatWithNewLine(text);
                    }
                } else {
                    RE.insertTextWithFormatWithNewLine(text);
                }
            } else {
                const range = document.createRange();
                range.selectNodeContents(RE.editor);
                range.collapse(false);
                sel.removeAllRanges();
                sel.addRange(range);
                document.execCommand("insertText", false, insertText);
            }
        }
    } else if (sel.rangeCount) {
        const range = sel.getRangeAt(0).cloneRange();
        range.deleteContents();
        const formattedText = document.createElement('span');
        formattedText.textContent = insertText;
        range.insertNode(formattedText);
        sel.removeAllRanges();
        sel.addRange(range);
    }

    RE.callback("input");
};

RE.insertTextWithFormatAtCurrentCursorPosition = function(text) {
    RE.restorerange();
    var sel = document.getSelection();
    if (sel.toString().length == 0) {
        document.execCommand("insertText", false, text);
    } else if (sel.rangeCount) {
        var range = sel.getRangeAt(0).cloneRange();
        range.deleteContents();
        range.insertNode(document.createTextNode(text));
        sel.removeAllRanges();
        sel.addRange(range);
    }
    RE.callback("input");
};

RE.insertTextWithFormatWithNewLine = function(text) {
    RE.restorerange();
    
    var editorContent = RE.editor.innerText.trim();
    var insertText = editorContent === "" ? text : "\n" + text;
    
    var sel = document.getSelection();
    if (sel.toString().length === 0) {
        
        RE.editor.focus();
        var range = document.createRange();
        range.selectNodeContents(RE.editor);
        range.collapse(false);
        sel.removeAllRanges();
        sel.addRange(range);
        document.execCommand("insertText", false, insertText);
    } else if (sel.rangeCount) {
        var range = sel.getRangeAt(0).cloneRange();
        range.deleteContents();
        range.insertNode(document.createTextNode(insertText));
        sel.removeAllRanges();
        sel.addRange(range);
    }
    RE.callback("input");
};


RE.insertImage = function(url, alt) {
    var img = document.createElement('img');
    img.setAttribute("src", url);
    img.setAttribute("alt", alt);
    img.onload = RE.updateHeight;

//    RE.insertHTML(img.outerHTML);
    RE.insertHTML(img.outerHTML + "<div><br /></div>");
    RE.callback("input");
    RE.blurFocus();
};

RE.insertEmoji = function(url, key) {
    var img = document.createElement('img');
    img.setAttribute("src", url);
    img.setAttribute("alt", key);
    img.setAttribute("emoji", key.slice(1, key.length-1));
    img.style.width = '12px';
    img.style.height = '12px';
    img.onload = RE.updateHeight;

    RE.insertHTML(img.outerHTML);
    //    RE.callback("input");
};

RE.setBlockquote = function() {
    document.execCommand('formatBlock', false, '<blockquote>');
};

RE.insertHTML = function(html) {
    RE.restorerange();
    document.execCommand('insertHTML', false, html);
};

RE.setAutoTextWrapping = function() {
    var style = document.createElement('style');
    style.innerHTML = '#editor { white-space: pre-wrap; word-wrap: break-word; }';
    document.head.appendChild(style);
};

RE.insertLink = function(url, title) {
    RE.restorerange();
    var sel = document.getSelection();
    if (sel.toString().length == 0) {
        document.execCommand("insertHTML",false,"<a href='"+url+"'>"+title+"</a>");
    } else if (sel.rangeCount) {
        var el = document.createElement("a");
        el.setAttribute("href", url);
        el.setAttribute("title", title);

        var range = sel.getRangeAt(0).cloneRange();
        range.surroundContents(el);
        sel.removeAllRanges();
        sel.addRange(range);
    }
    RE.callback("input");
};

RE.insertAttachment = function(url, thumbnailUrl) {
    var html = `<div><br><a class="attachment" href="` + url + `"><img src="` + thumbnailUrl + `"/></a></div>`;
    //Note: Inserting a break explicity here to let editor scroll to the bottom automatically.
    RE.insertHTML(html + "<div><br /></div>");
    RE.callback("input");
};

RE.prepareInsert = function() {
    RE.backuprange();
};

RE.backuprange = function() {
    var selection = window.getSelection();
    if (selection.rangeCount > 0) {
        var range = selection.getRangeAt(0);
        RE.currentSelection = {
            "startContainer": range.startContainer,
            "startOffset": range.startOffset,
            "endContainer": range.endContainer,
            "endOffset": range.endOffset
        };
    }
};

RE.addRangeToSelection = function(selection, range) {
    if (selection) {
        selection.removeAllRanges();
        selection.addRange(range);
    }
};

// Programatically select a DOM element
RE.selectElementContents = function(el) {
    var range = document.createRange();
    range.selectNodeContents(el);
    var sel = window.getSelection();
    // this.createSelectionFromRange sel, range
    RE.addRangeToSelection(sel, range);
};

RE.restorerange = function() {
    var selection = window.getSelection();
    selection.removeAllRanges();
    var range = document.createRange();
    range.setStart(RE.currentSelection.startContainer, RE.currentSelection.startOffset);
    range.setEnd(RE.currentSelection.endContainer, RE.currentSelection.endOffset);
    selection.addRange(range);
};

RE.focus = function() {
    var range = document.createRange();
    range.selectNodeContents(RE.editor);
    range.collapse(false);
    var selection = window.getSelection();
    selection.removeAllRanges();
    selection.addRange(range);
    RE.editor.focus();
};

//RE.focusAtPoint = function(x, y) {
//    var range = document.caretRangeFromPoint(x, y);
//    var selection = window.getSelection();
//    selection.removeAllRanges();
//    selection.addRange(range);
//    RE.editor.focus();
//};

RE.focusAtEnd = function() {
    var range = document.caretRangeFromPoint(RE.editor.offsetLeft + RE.editor.clientWidth - 1, RE.editor.offsetTop + RE.editor.clientHeight - 1);
    var selection = window.getSelection();
    selection.removeAllRanges();
    selection.addRange(range);
    RE.editor.focus();
};

RE.focusAtPoint = function(x, y) {
    var range;
    if(x > RE.editor.offsetLeft &&
       x < RE.editor.offsetLeft + RE.editor.clientWidth &&
       y > RE.editor.offsetTop &&
       y < RE.editor.offsetTop + RE.editor.clientHeight){
        range = document.caretRangeFromPoint(x, y);
    }else{
        //range = document.caretRangeFromPoint(RE.editor.offsetLeft + 1, RE.editor.offsetTop + 1);
        range = document.caretRangeFromPoint(RE.editor.offsetLeft + RE.editor.clientWidth - 1, RE.editor.offsetTop + RE.editor.clientHeight - 1);
    }
    var selection = window.getSelection();
    selection.removeAllRanges();
    selection.addRange(range);
    RE.editor.focus();
};

RE.blurFocus = function() {
    RE.editor.blur();
};

/**
Recursively search element ancestors to find a element nodeName e.g. A
**/
var _findNodeByNameInContainer = function(element, nodeName, rootElementId) {
    if (element.nodeName == nodeName) {
        return element;
    } else {
        if (element.id === rootElementId) {
            return null;
        }
        _findNodeByNameInContainer(element.parentElement, nodeName, rootElementId);
    }
};

var isAnchorNode = function(node) {
    return ("A" == node.nodeName);
};

RE.getAnchorTagsInNode = function(node) {
    var links = [];

    while (node.nextSibling !== null && node.nextSibling !== undefined) {
        node = node.nextSibling;
        if (isAnchorNode(node)) {
            links.push(node.getAttribute('href'));
        }
    }
    return links;
};

RE.getAnchorTagsTitleInNode = function(node) {
    var links = [];

    while (node.nextSibling !== null && node.nextSibling !== undefined) {
        node = node.nextSibling;
        if (isAnchorNode(node)) {
            links.push(node.getAttribute('title'));
        }
    }
    return links;
};

RE.countAnchorTagsInNode = function(node) {
    return RE.getAnchorTagsInNode(node).length;
};

/**
 * If the current selection's parent is an anchor tag, get the href.
 * @returns {string}
 */
RE.getSelectedHref = function() {
    var href, sel;
    href = '';
    sel = window.getSelection();
    if (!RE.rangeOrCaretSelectionExists()) {
        return null;
    }

    var tags = RE.getAnchorTagsInNode(sel.anchorNode);
    //if more than one link is there, return null
    if (tags.length > 1) {
        return null;
    } else if (tags.length == 1) {
        href = tags[0];
    } else {
        var node = _findNodeByNameInContainer(sel.anchorNode.parentElement, 'A', 'editor');
        href = node.href;
    }

    return href ? href : null;
};

RE.getSelectedHrefTitle = function() {
    var href, sel;
    href = '';
    sel = window.getSelection();
    if (!RE.rangeOrCaretSelectionExists()) {
        return null;
    }

    var tags = RE.getAnchorTagsInNode(sel.anchorNode);
    //if more than one link is there, return null
    if (tags.length > 1) {
        return null;
    } else if (tags.length == 1) {
        href = tags[0];
    } else {
        var node = _findNodeByNameInContainer(sel.anchorNode.parentElement, 'A', 'editor');
        href = node.innerText;
    }

    return href ? href : null;
};

RE.getSelectedText = function() {
    var text, sel;
    text = '';
    sel = window.getSelection();
    if (!RE.rangeOrCaretSelectionExists()) {
        return null;
    }
    text = sel.toString();
    return text;
}

// Returns the cursor position relative to its current position onscreen.
// Can be negative if it is above what is visible
RE.getRelativeCaretYPosition = function() {
    var y = 0;
    var sel = window.getSelection();
    if (sel.rangeCount) {
        var range = sel.getRangeAt(0);
        var needsWorkAround = (range.startOffset == 0)
        /* Removing fixes bug when node name other than 'div' */
        // && range.startContainer.nodeName.toLowerCase() == 'div');
        if (needsWorkAround) {
            y = range.startContainer.offsetTop - window.pageYOffset;
        } else {
            if (range.getClientRects) {
                var rects=range.getClientRects();
                if (rects.length > 0) {
                    y = rects[0].top;
                }
            }
        }
    }

    return y;
};

// Returns the cursor position relative to its current position onscreen.
// Can be negative if it is above what is visible
RE.getRelativeCaretXPosition = function() {
    var x = 0;
    var sel = window.getSelection();
    if (sel.rangeCount) {
        var range = sel.getRangeAt(0);
        var needsWorkAround = (range.startOffset == 0)
        /* Removing fixes bug when node name other than 'div' */
        // && range.startContainer.nodeName.toLowerCase() == 'div');
        if (needsWorkAround) {
            x = range.startContainer.offsetLeft - window.pageXOffset;
        } else {
            if (range.getClientRects) {
                var rects=range.getClientRects();
                if (rects.length > 0) {
                    x = rects[0].left;
                }
            }
        }
    }

    return x;
};

RE.getSelectionStartX = function() {
    var x = 0;
    var sel = window.getSelection();
    if (sel.rangeCount) {
        var range = sel.getRangeAt(0);
        if (range.getClientRects) {
            var rects=range.getClientRects();
            if (rects.length > 0) {
                x = rects[0].left;
            }
        }
    }

    return x;
};

RE.getSelectionStartY = function() {
    var x = 0;
    var sel = window.getSelection();
    if (sel.rangeCount) {
        var range = sel.getRangeAt(0);
        if (range.getClientRects) {
            var rects=range.getClientRects();
            if (rects.length > 0) {
                x = rects[0].top;
            }
        }
    }

    return x;
};

RE.deleteSelection = function() {
    var sel = window.getSelection();
    sel.deleteFromDocument();
};
